/*
 *  _common.c
 *
 *  Implementation of common routines
 *  for creating Notes utilities using C and NotesAPI
 *
 *  Written by Ilya Evseev (evseev@csa.ru) at Oct 2000
 *
 *  See detailed documented example of using NSFSearch in Notes API SDK:
 *  \SAMPLES\ADMIN\FIND_DBS\FIND_DBS.C
 *
 *  How to build sample for testing do_action_for_each_database() routine:
 *   bcc32 -Ie:\NotesAPI.462\Include -DTEST_DB_ENUM -eTestEnum _common.c nnotes_bc.lib
 */

#include "_common.h"

#include <osmisc.h>     /* OSLoadString */
#include <ods.h>        /* ODSLength */
#include <osfile.h>     /* OSPathNetConstruct, FILE_xxx */
#include <oserr.h>      /* ERR_NOEXIST */
#include <nsfdb.h>      /* NSFDbOpen, NSFDbClose, NSFDb... */

#if defined(_WIN32)
#  if defined(__BORLANDC__)
#    pragma comment(lib,"nnotes_bc.lib")
#  endif
#  if defined(__WIN32_CONSOLE__)
#    include <windows.h>   /* CharToOemA */
#    if defined(_MSC_VER)
#      pragma comment(lib,"user32.lib")
#    endif
#  endif
#endif


/*-------------------------------------------------------------
 *  GNU C don't have strlwr/strupr
 */
#if defined(__GNUC__)
 
#include <ctype.h>

char *strlwr(char * const s)
{
    if (s) {
	char *p;
	for (p = s; *p; p++) {
	    *p = _tolower(*p);
	}
    }
    return s;
} 
 
#endif /* if defined(__GNUC__) */


/*-------------------------------------------------------------
 *  notes_perror
 */
int notes_perror(STATUS status, const char *where)
{
    if (status != NOERROR) { 
        char error_text_lotus[ERROR_TEXT_SIZE];
        char error_text_native[ERROR_TEXT_SIZE];
        OSLoadString(NULLHANDLE, ERR(status), error_text_lotus, sizeof(error_text_lotus)-1);
	//fprintf(STDERR, "\nRAW: \"%s\"\n", error_text_lotus);

#   if defined(OS390) && (__STRING_CODE_SET__ != ISO8859-1) /* ebcdic compile */
        OSLMBCSToNative(error_text_lotus, MAXWORD,
            error_text_native, sizeof(error_text_native));
#   else  /* !if defined(OS390, ebcdic compile) */
        OSTranslate(OS_TRANSLATE_LMBCS_TO_NATIVE,
            error_text_lotus, strlen(error_text_lotus),
            error_text_native, sizeof(error_text_native)-1);

#     if defined(__WIN32_CONSOLE__)
        CharToOemA(error_text_native, error_text_native);
#     endif

#   endif
        fprintf(STDERR, "\n%s: error %d: %s\n", where ? where : "???",
            status, error_text_native);
    }
    return status;
} 


/*-------------------------------------------------------------
 *  get_summary_item
 */
void *get_summary_item(ITEM_TABLE *summary, char *item_name,
    USHORT *ptype, int *psize)
{
    int i, nitems, namelen = strlen(item_name);
    USHORT *psizes, *pnamesize, *pvalsize;
    char *pvalues;
    char *name;
    USHORT datatype;
    void *value;

    if (!summary || !item_name)
        return NULL;

    nitems = summary->Items;
    psizes = (USHORT*)( ((char*)summary) + ODSLength(_ITEM_TABLE) );
    pvalues = ((char*)psizes) + nitems * sizeof(USHORT) * 2;

    pnamesize = psizes;
    name = pvalues;
    for (i = 0; i < nitems; i++) {
        pvalsize = pnamesize + 1;
        datatype = *(USHORT*)(name + *pnamesize);
        value = name + *pnamesize + sizeof(USHORT);
        if ((!ptype || (!*ptype || *ptype == datatype)) &&
            (*pnamesize == namelen) &&
            !memcmp(name, item_name, namelen))
        {
            if (ptype)
                *ptype = datatype;
            if (psize)
                *psize = *pvalsize - sizeof(USHORT);
            return value;
        }
        name += *pnamesize + *pvalsize;
        pnamesize += 2;
    }
    return NULL;
}


/*-------------------------------------------------------------
 *  get_summary_textitem
 *  get_summary_filepath
 */
char *get_summary_textitem(char *dest_str, ITEM_TABLE *summary, char *item_name)
{
    USHORT dtype = TYPE_TEXT;
    int len;
    void *p = get_summary_item(summary, item_name, &dtype, &len);
    if (!p)
        return NULL;
    memcpy(dest_str, p, len);
    dest_str[len] = 0;
    return dest_str;
}

char *get_summary_filepath(char *dest_str, ITEM_TABLE *summary)
{
    return get_summary_textitem(dest_str, summary, "$Path");
}


/*-------------------------------------------------------------
 *  dblist_enumerator
 */

typedef struct {
    STATUS (*action)(HANDLE db, ITEM_TABLE *summary, void *args);
    void *action_args;
    char *server;
} Enumerator_args;

static STATUS LNPUBLIC dblist_enumerator(void *enumerator_args_ptr,
    SEARCH_MATCH *SearchMatch, ITEM_TABLE *summary)
{
    Enumerator_args *enumerator_args = (Enumerator_args *)enumerator_args_ptr;
    STATUS  status;
    HANDLE  db;
    USHORT  db_mode;
    char    filepath[MAXPATH];
    char    fullpath[MAXPATH];
    static char empty[] = "";

    if ((SearchMatch->SERetFlags & SE_FMATCH) == 0)
        return NOERROR;

    if (!get_summary_textitem(filepath, summary, "$Path"))
        goto bad_summary;  // status = ERR_NOEXIST;

    IF_ERROR("DirSearch..OSPathNetConstruct",
        OSPathNetConstruct(
            NULL,
            enumerator_args->server ? enumerator_args->server : empty,
            filepath,
            fullpath))
        goto bad_path;

    IF_ERROR("DirSearch..NSFDbOpen", NSFDbOpen(fullpath, &db)) {
        printf("(%s)\n", fullpath);
        goto bad_open;
    }
    IF_ERROR("DirSearch..NSFDbGetMode", NSFDbModeGet(db, &db_mode))
        goto bad_mode;

    if ((db_mode & DB_DIRECTORY) != 0) {
        char dbpath[MAXPATH];
        IF_ERROR("DirSearch..NSFDbPathGet", NSFDbPathGet(db, NULL, dbpath))
            goto bad_dir;
        status = do_action_for_each_database(dbpath, NULL, 
            enumerator_args->action, enumerator_args->action_args);
    } else {
        status = enumerator_args->action(db, summary, enumerator_args->action_args);
        NSFDbClose(db);
    }
    return status;

bad_dir:
bad_mode:
    NSFDbClose(db);
bad_open:
bad_path:
bad_summary:
    return NOERROR;  //don't stop enumeration by erroneous 'status'
}


/*-------------------------------------------------------------
 *  do_action_for_each_database
 */
STATUS do_action_for_each_database(char *filename, char *server,
    STATUS (*action)(HANDLE db, ITEM_TABLE *summary, void *args),
    void *action_args)
{
    STATUS status, e;
    char fullpath[MAXPATH];
    USHORT mode;
    HANDLE db;
    Enumerator_args enumerator_args;

    if (server) {
        RETURN_IF_ERROR("DirSearch.OSPathNetConstruct",
            OSPathNetConstruct(NULL, server, filename, fullpath));
    } else {
        static char servname[MAXPATH];
        strcpy(fullpath, filename);
        RETURN_IF_ERROR("DirSearch.OSPathNetParse",
            OSPathNetParse(fullpath, NULL, server = servname, NULL));
    }
    
    IF_ERROR("DirSearch.NSFDbOpen", NSFDbOpen(fullpath, &db)) {
        printf("(%s)\n", fullpath);
        goto bad_open;
    }

    GOTO_IF_ERROR(bad_db, "DirSearch.NSFDbModeGet", NSFDbModeGet(db, &mode));

    if (mode & DB_LOADED) {

        struct {
            ITEM_TABLE header;
            USHORT sizes[2];
            char name_dtype_value[1024];
        } summary;
        static char Path[] = "$Path";

        summary.header.Items = 1;
        summary.sizes[0] = sizeof(Path)-1;
        summary.sizes[1] = sizeof(USHORT) + strlen(fullpath);
        summary.header.Length = sizeof(summary) +
            summary.sizes[0] + summary.sizes[1];
        memcpy(summary.name_dtype_value, Path, sizeof(Path)-1);
        *(USHORT*)(summary.name_dtype_value + sizeof(Path)-1) = TYPE_TEXT;
        memcpy(summary.name_dtype_value + sizeof(Path)-1 + sizeof(USHORT),
            fullpath, strlen(fullpath));

        status = action(db, &summary.header, action_args);

    } else if (mode & DB_DIRECTORY) {

        enumerator_args.action = action;
        enumerator_args.action_args = action_args;
        enumerator_args.server = server;

        GOTO_IF_ERROR(bad_db, "DirSearch.NSFSearch",
            NSFSearch (
                db,                /* really directory handle */
                NULLHANDLE,        /* selection formula */
                NULL,              /* title of view in formula */
                SEARCH_FILETYPE +  /* search for files */
                SEARCH_SUMMARY,    /* return a summary buffer */
                FILE_DBANY +       /* find any .NS? file */
                FILE_DIRS +        /* find subdirectories */
                FILE_NOUPDIRS,     /* don't find the ".." dir */
                NULL,              /* starting date */
                dblist_enumerator, /* call for each file found */
                &enumerator_args,  /* argument to action routine */
                NULL));            /* returned ending date (unused) */
    } else {
        status = ERR_INV_FILE_NAME;   /* what can i do ??? */
    }

bad_db:
    e = NSFDbClose(db);
    if (status == NOERROR)
        status = e;
bad_open:
    return status;
}


void terminate_notes_session(void) {
    NotesTerm();
}


/*-------------------------------------------------------------
 *  TESTING PART
 */

// #define TEST_PERROR 1

#if defined(TEST_DB_ENUM)

#include <nsferr.h>

STATUS action(HANDLE db, ITEM_TABLE *summary, void *p)
{
    char db_info[NSF_INFO_SIZE];
    char db_filepath[MAXPATH];
    char canonical[MAXPATH];
    char expanded[MAXPATH];
    STATUS status;
    USHORT dbmode;

    RETURN_IF_ERROR("NSFDbModeGet", NSFDbModeGet(db, &dbmode));
    RETURN_IF_ERROR("NSFDbPathGet", NSFDbPathGet(db, canonical, expanded));
    if ((dbmode & DB_DIRECTORY) != 0) {
        strcpy(db_info, "(-Directory-)");
        do_action_for_each_database(expanded, NULL, action, p);
    } else {
        RETURN_IF_ERROR("NSFDbInfoGet", NSFDbInfoGet(db, db_info));
    }
    printf("%s\ncanonical: %s\n expanded: %s\n filepath: %s\n   dbinfo: %s\n",
        p,
        canonical,
        expanded,
        get_summary_filepath(db_filepath, summary),
        db_info);
    return NOERROR;
}

int main(int argc, char *argv[])
{
    STATUS status = NOERROR;

    if (argc == 2 || argc == 3) {
        INIT_NOTES_SESSION();
        status = do_action_for_each_database(
            argv[1], (argc > 2 ? argv[2] : NULL), action, " --=< Huh >=-- ");
    } else {
        printf("Usage: %s dirname|dbname [server]\n", argv[0]);
    }
    return status;
}

#elif defined(TEST_PERROR)  /* if defined(TEST_DB_ENUM) */

#include <stdlib.h>

int main(int argc, char *argv[])
//STATUS NotesMain(int argc, char *argv[])
//LAPI_MAIN
{
    STATUS status;
    int i;
    char s[16];

    printf("Interpret Notes error codes:\n");

    INIT_NOTES_SESSION();

    for (i = 1; i < argc; i++) {
        sprintf(s, "%d", i);  // Linux don't understand itoa(i, s, 10)
        notes_perror(atoi(argv[i]), s);
    }

    return status;
}

#elif defined(TEST_TRANSLATE)  /* elif defined(TEST_PERROR) */

//#include <osmisc.h>

#pragma argsused
int main(int argc, char *argv[])
{
    char buf[1024];
    char text[] = "   Win1251  --  Cyrillic text encoded in Win1251 codepage";
    int buflen;
    STATUS status;
    
    INIT_NOTES_SESSION();

    buflen = OSTranslate(
        OS_TRANSLATE_NATIVE_TO_LMBCS,
        text,
        strlen(text)+1,
        buf,
        //1);
        sizeof(buf)-1);
    printf("buflen = %d\n", buflen);
    printf("buf = \"%s\"\n", buf);
    return 0;
}

#endif  /* elif defined(TEST_TRANSLATE) */

/* EOF */
