/*
 *  lnacls.c
 *
 *  Dump and edit ACL (access control list) for Lotus Notes database(s)
 *
 *  See syntax() routine or run without arguments for all using details
 *
 *  Written by Ilya Evseev (evseev@csa.ru) at Feb 2001
 */

#include <string.h>       // ANSI C string routines
#include <stdlib.h>       // getenv()

#include "_common.c"      // my own file for LN-aware programming
#include <dname.h>        // DNCanonicalize()
#include <nsfdb.h>        // NSFDbReadACL(),...
#include <acl.h>
#include <osmem.h>        // OSMemFree(acl)
#include <miscerr.h>      // ERR_ACL_NOENTRY


char *server_name;


static char sCanCreateDocuments[]     = "cd";
static char sCanDeleteDocuments[]     = "dd";
static char sCanCreatePersonalAgent[] = "cpa";
static char sCanCreatePersonalFolder[]= "cpf";
static char sCanCreateSharedFolders[] = "csf";
static char sCanCreateLotusScript[]   = "cls";
static char sPublicReader[]           = "rpd";
static char sPublicWriter[]           = "wpd";

static char mod_prefix[]  = " ";
static char role_prefix[] = " ";


char *extract_token(char **pp, char delim)
{
#  define p (*pp)

    char *result;
    if (!pp || !*pp || !**pp)
        return "";
    result = p;
    for (; *p; p++) {
        if (*p == delim) {
            *p++ = '\0';
            break;
        }
    }
    return result;

#  undef p
}


STATUS acl_get_roles(HANDLE acl, char roles[ACL_PRIVCOUNT][ACL_PRIVSTRINGMAX])
{
    STATUS status;
    WORD i;
    for (i = 0; i < ACL_PRIVCOUNT - 1; i++) {
        status = ACLGetPrivName(acl, i, roles[i]);
        if (status == ERR_ACL_NOENTRY) {
            if (i >= ACL_BITPRIVCOUNT) {
                status = NOERROR;
                break;
            }
        } else IF_ERROR("GetRoles.ACLGetPrivName", status) {
            break;
        }
    }
    return status;
}


STATUS set_user_type (WORD *user_type, char *utype_str, WORD *update_mode_ptr)
{
#define SET_FLAGS_IF(_Str, _Enable, _Disable)  \
        if (!strncmp(_Str, utype_str, utype_len)) {  \
            *user_type |= (_Enable);  \
            *user_type &= ~(_Disable);  \
            *update_mode_ptr |= ACL_UPDATE_FLAGS;  \
        }

    int utype_len;

    if (!utype_str || !*utype_str || !user_type || !update_mode_ptr)
        return 0;

    utype_len = strlen(utype_str);

         SET_FLAGS_IF( "person", ACL_FLAG_PERSON, ACL_FLAG_GROUP | ACL_FLAG_SERVER)
    else SET_FLAGS_IF( "server", ACL_FLAG_SERVER, ACL_FLAG_GROUP | ACL_FLAG_PERSON)
    else SET_FLAGS_IF(  "group", ACL_FLAG_GROUP,  ACL_FLAG_PERSON | ACL_FLAG_SERVER)
    else SET_FLAGS_IF( "pgroup", ACL_FLAG_GROUP | ACL_FLAG_PERSON, ACL_FLAG_SERVER)
    else SET_FLAGS_IF( "sgroup", ACL_FLAG_GROUP | ACL_FLAG_SERVER, ACL_FLAG_PERSON)
    else SET_FLAGS_IF("unknown", 0, ACL_FLAG_PERSON | ACL_FLAG_SERVER | ACL_FLAG_GROUP)
    else return notes_perror(ERR_TEXT_PATTERN, "SetUserType");

    return NOERROR;

#undef SET_FLAGS_IF
}


STATUS set_access_level (WORD *level, char *level_str, WORD *update_mode_ptr)
{
#define SET_LEVEL_IF(_Str, _Level) \
        if (!strncmp(_Str, level_str, level_len)) { \
            *level = _Level;  \
            *update_mode_ptr |= ACL_UPDATE_LEVEL;  \
        }

    int level_len;

    if (!level_str || !*level_str || !level) {
        return NOERROR;
    }

    level_len = strlen(level_str);

         SET_LEVEL_IF( "noaccess", ACL_LEVEL_NOACCESS)
    else SET_LEVEL_IF("depositor", ACL_LEVEL_DEPOSITOR)
    else SET_LEVEL_IF(   "reader", ACL_LEVEL_READER)
    else SET_LEVEL_IF(   "author", ACL_LEVEL_AUTHOR)
    else SET_LEVEL_IF(   "editor", ACL_LEVEL_EDITOR)
    else SET_LEVEL_IF( "designer", ACL_LEVEL_DESIGNER)
    else SET_LEVEL_IF(  "manager", ACL_LEVEL_MANAGER)
    else return notes_perror(ERR_TEXT_PATTERN, "SetAccessLevel");

    return NOERROR;
}


STATUS set_modifiers(WORD *flags, char *customs, bool enable, WORD *update_mode_ptr)
{
#define SET_FLAGS_IF(_Str, _Flags)  \
        if (!strcmp(p, _Str)) {  \
            if (enable) {  \
                *flags |= (_Flags);  \
            } else {  \
                *flags &= ~(_Flags);  \
            }  \
            *update_mode_ptr |= ACL_UPDATE_FLAGS;  \
        }

#define CLEAR_FLAGS_IF(_Str, _Flags)  \
        if (!strcmp(p, _Str)) {  \
            if (!enable) {  \
                *flags |= (_Flags);  \
            } else {  \
                *flags &= ~(_Flags);  \
            }  \
            *update_mode_ptr |= ACL_UPDATE_FLAGS;  \
        }

    char *p;

    if (!customs || !*customs || !flags || !update_mode_ptr)
        return NOERROR;

    for (p = strtok(customs,"+"); p; p = strtok(NULL,"+")) {
             CLEAR_FLAGS_IF(sCanCreateDocuments   , ACL_FLAG_AUTHOR_NOCREATE)
        else CLEAR_FLAGS_IF(sCanDeleteDocuments   , ACL_FLAG_NODELETE)
        else SET_FLAGS_IF(sCanCreatePersonalAgent , ACL_FLAG_CREATE_PRAGENT)
        else SET_FLAGS_IF(sCanCreatePersonalFolder, ACL_FLAG_CREATE_PRFOLDER)
        else SET_FLAGS_IF(sCanCreateSharedFolders , ACL_FLAG_CREATE_FOLDER)
        else SET_FLAGS_IF(sCanCreateLotusScript   , ACL_FLAG_CREATE_LOTUSSCRIPT)
        else SET_FLAGS_IF(sPublicReader           , ACL_FLAG_PUBLICREADER)
        else SET_FLAGS_IF(sPublicWriter           , ACL_FLAG_PUBLICWRITER)
        else return notes_perror(ERR_TEXT_PATTERN, "SetModifiers");
    }
    return NOERROR;
}


typedef struct {
    char **data;
    int count;
} Acl_modifiers;


typedef struct {
    char *name;
    WORD access_level;
    ACL_PRIVILEGES *privileges;
    WORD access_flags;
} Acl_entry;


typedef struct {
    int count;
    int total;
    Acl_entry data[1];
} Acl_entries;


#if defined(__BORLANDC__)
#  pragma argsused
#endif

void LNCALLBACK count_acl_entries(void *EnumFuncParam,
                       char *name,
                       WORD access_level,
                       ACL_PRIVILEGES *privileges,
                       WORD access_flags)
{
    (*((int*)EnumFuncParam))++;
}


void LNCALLBACK list_acl_entries(void *EnumFuncParam,
                      char *name,
                      WORD access_level,
                      ACL_PRIVILEGES *privileges,
                      WORD access_flags)
{
     Acl_entries *elist = (Acl_entries*)EnumFuncParam;
     Acl_entry   *e     = &elist->data[elist->count++];
     e->name            = name;
     e->access_level    = access_level;
     e->privileges      = privileges;
     e->access_flags    = access_flags;
}


STATUS acl_clear_all(HANDLE acl, Acl_entries *acl_entries)
{
     STATUS status;
     int i;
     for (i = 0; i < acl_entries->count; i++) {
         IF_ERROR("EditACL.DeleteAllEntries",
             ACLDeleteEntry(acl, acl_entries->data[i].name)
         );
     }
     acl_entries->count = 0;
     return status;
}


int acl_role_getnum(char acl_roles[ACL_PRIVCOUNT][ACL_PRIVSTRINGMAX], char *p)
{
    int i, len, qlen;
    char *q;
    if (!p || !*p)
        return -1;
    if (*p == '[')
        p++;
    len = strlen(p);
    if (p[len-1] == ']')
        len--;
    for (i = 0; i < ACL_PRIVCOUNT; i++) {
        q = acl_roles[i];
        if (!*q)
            continue;
        if (*q == '[')
            q++;
        qlen = strlen(q);
        if (q[qlen-1] == ']')
            qlen--;
        if (len == qlen && !memcmp(p, q, len))
            return i;
    }
    return -1;  //not found
}


#if defined(__BORLANDC__)
#  pragma argsused  /* 'summary' is not used */
#endif

STATUS edit_acl_by_dbhandle(HANDLE db, ITEM_TABLE *summary, void *args)
{
    HANDLE         acl;
    bool           is_acl_changed = false;
    STATUS         status;
    char           dbpath[MAXPATH];
    Acl_modifiers *acl_modifiers = (Acl_modifiers*)args;
    Acl_entries   *acl_entries = NULL;
    int            nacl_entries = 0;
    char           acl_roles[ACL_PRIVCOUNT][ACL_PRIVSTRINGMAX];
    int i, j;

    RETURN_IF_ERROR("EditACL.NSFDbPathGet", NSFDbPathGet(db, NULL, dbpath));
    printf("%s...", dbpath);

    IF_ERROR("EditACL.NSFDbReadACL", NSFDbReadACL(db, &acl))
        goto bad_acl;

    IF_ERROR("EditACL.GetRoles", acl_get_roles(acl, acl_roles))
        goto bad_roles;

    for (i = 0; i < acl_modifiers->count; i++) {
        Acl_entry *acle;

        bool new_state = true;
        bool erase_before_edit = false;
        WORD update_mode = 0;

        char *who, *utype, *level, *customs, *roles;
        char who_name[MAXUSERNAME];

        char curr_mod[1024];
        char *p = curr_mod;
        strcpy(curr_mod, acl_modifiers->data[i]);
	
        IF_ERROR("EditACL.CountACLEntries", ACLEnumEntries(acl, count_acl_entries, &nacl_entries))
            goto bad_count;

        if ((acl_entries = (Acl_entries*)malloc(sizeof(Acl_entries) + (nacl_entries - 1) * sizeof(Acl_entry))) == NULL) {
            status = notes_perror(ERR_MEMORY, "EditACL.malloc");
            goto bad_malloc;
        }
        acl_entries->count = 0;
        acl_entries->total = nacl_entries;

        IF_ERROR("EditACL.ListEntries", ACLEnumEntries(acl, list_acl_entries, acl_entries))
            goto bad_list;

        if (*p == '!') {
            acl_clear_all(acl, acl_entries);
            is_acl_changed = true;
            if (!*++p)   // modifier is empty?
                continue;
        }

        switch (*p++) {
            case '+':  break;
            case '-':  new_state = false;  break;
            case '=':  erase_before_edit = true;  break;
            default:   status = notes_perror(ERR_BADQUOTE_FORMAT, "EditACL.CmdChar");
                       goto bad_cmd;
        }

        who     =        extract_token(&p, ':');
        utype   = strlwr(extract_token(&p, ':'));
        level   = strlwr(extract_token(&p, ':'));
        customs = strlwr(extract_token(&p, ':'));
        roles   =        extract_token(&p, ':');

        IF_ERROR("EditACL.DNCanonicalize",
            DNCanonicalize(0L, NULL, who, who_name, ELEMS(who_name), NULL))
                goto bad_name;

        //  find entry, erase if need, create if not exist
        //
        for (acle = acl_entries->data, j = 0; j < acl_entries->count; acle++, j++) {
            int is_default = (!who_name || !*who_name) && !strcmp(acle->name, ACL_DEFAULT_NAME);
            if (is_default || !strcmp(acle->name, who_name))
            {
                if (!is_default && erase_before_edit) {
                    IF_ERROR("EditACL.ACLDeleteEntry", ACLDeleteEntry(acl, acle->name)) {
                    } else {
                        --j;
                        --acl_entries->count;
                    }
                } else {
                    break;
                }
            }
        }
        if (j == acl_entries->count) {
            static Acl_entry a;
            static ACL_PRIVILEGES priv;

            if (!new_state)   // entry must be deleted and is not found
                continue;     // ...then process next entry

            memset(&priv, 0, sizeof(priv));
            a.name = who_name;
            a.access_level = ACL_LEVEL_NOACCESS;
            a.privileges = &priv;
            a.access_flags = 0;
            acle = &a;
            update_mode = (WORD)-1;
        }

        IF_ERROR("EditACL.SetUserType",    set_user_type(&acle->access_flags, utype, &update_mode))
            goto bad_utype;
        IF_ERROR("EditACL.SetAccessLevel", set_access_level(&acle->access_level, level, &update_mode))
            goto bad_level;
        IF_ERROR("EditACL.SetModifiers",   set_modifiers(&acle->access_flags, customs, new_state, &update_mode))
            goto bad_mods;

        //  set roles
        //
        if (*roles) {
            char *p;
            for (p = strtok(roles,"+"); p; p = strtok(NULL,"+")) {
                int n = acl_role_getnum(acl_roles, p);
                if (n >= 0) {
                    if (new_state) {
                        ACLSetPriv(*acle->privileges, n);
                    } else {
                        ACLClearPriv(*acle->privileges, n);
                    }
                    update_mode |= ACL_UPDATE_PRIVILEGES;
                }
            }
        }

        //  if no partial deletions, entry must be deleted entirely
        //
        if (!new_state && !update_mode) {
            IF_ERROR("EditACL.ACLDeleteEntry", ACLDeleteEntry(acl, acle->name))
                goto bad_delete;
            is_acl_changed = true;
        }

        if (update_mode) {
            // NULL will be treated as "-Default-"
            char *acle_name = (!*acle->name || !strcmp(acle->name, ACL_DEFAULT_NAME)) ? NULL : acle->name;
            is_acl_changed = true;
            status = ACLUpdateEntry(
                     acl,
                     acle_name,
                     update_mode,
                     NULL,
                     acle->access_level,
                     acle->privileges,
                     acle->access_flags);
            if (status == ERR_ACL_NOENTRY) {
                status = ACLAddEntry(
                         acl,
                         acle_name,
                         acle->access_level,
                         acle->privileges,
                         acle->access_flags);
            }
            IF_ERROR("EditACL.ACLUpdateEntry", status)
                goto bad_update;
        }

        free(acl_entries);
        acl_entries = NULL;

    } // foreach(acl_modifiers)

    printf(" OK\n");

    if (is_acl_changed)
        IF_ERROR("EditACL.NSFDbStoreACL", NSFDbStoreACL(db, acl, 0L, 0));

bad_update:
bad_mods:
bad_level:
bad_utype:
bad_delete:
bad_name:
bad_cmd:
bad_list:
    if (acl_entries)
        free(acl_entries);
bad_malloc:
bad_count:
    OSMemFree(acl);
bad_roles:
bad_acl:
    return 0;  //don't break enumeration by 'status';
}


STATUS edit_acl_by_filename(char *filename, char **acl_modifiers, int nmodifiers)
{
    Acl_modifiers a;
    a.data = acl_modifiers;
    a.count = nmodifiers;
    return do_action_for_each_database(filename, server_name,
        edit_acl_by_dbhandle, &a);
}


const char *present_or_blank (bool is_present, const char *present_str)
{
    static char buff[] = "                                           ";
    if (is_present)
        return present_str;
    return buff + sizeof(buff) - 1 - strlen(present_str);
}


void LNPUBLIC dump_acl_entry(void *roles_ptr,    // Additional Args
                             char *entry_name,
                             WORD access_level,
                             ACL_PRIVILEGES *privileges,
                             WORD access_flags)
{
    char *role;
    char *access_str;
    int   i;

    //  display roles
    //
    if (!getenv("NO_ROLES")) {
        for (role = (char*)roles_ptr, i = 0; i < ACL_PRIVCOUNT; role += ACL_PRIVSTRINGMAX, i++) {
            if (*role) {
                printf("%s%s", role_prefix, present_or_blank(ACLIsPrivSet(*privileges, i), role));
            } else if (i >= ACL_BITPRIVCOUNT) {
                break;
            }
        }
        printf("  ");
    }

    //  display modifiers
    //
    printf("%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
        mod_prefix, present_or_blank(!(access_flags & ACL_FLAG_AUTHOR_NOCREATE)  , sCanCreateDocuments),
        mod_prefix, present_or_blank(!(access_flags & ACL_FLAG_NODELETE)         , sCanDeleteDocuments),
        mod_prefix, present_or_blank(  access_flags & ACL_FLAG_CREATE_PRAGENT    , sCanCreatePersonalAgent),
        mod_prefix, present_or_blank(  access_flags & ACL_FLAG_CREATE_PRFOLDER   , sCanCreatePersonalFolder),
        mod_prefix, present_or_blank(  access_flags & ACL_FLAG_CREATE_FOLDER     , sCanCreateSharedFolders),
        mod_prefix, present_or_blank(  access_flags & ACL_FLAG_CREATE_LOTUSSCRIPT, sCanCreateLotusScript),
        mod_prefix, present_or_blank(  access_flags & ACL_FLAG_PUBLICREADER      , sPublicReader),
        mod_prefix, present_or_blank(  access_flags & ACL_FLAG_PUBLICWRITER      , sPublicWriter)
        );

    //  display name and access level
    //
         if (access_level == ACL_LEVEL_NOACCESS)  access_str = "NoAccess ";
    else if (access_level == ACL_LEVEL_DEPOSITOR) access_str = "Depositor";
    else if (access_level == ACL_LEVEL_READER)    access_str = "Reader   ";
    else if (access_level == ACL_LEVEL_AUTHOR)    access_str = "Author   ";
    else if (access_level == ACL_LEVEL_EDITOR)    access_str = "Editor   ";
    else if (access_level == ACL_LEVEL_DESIGNER)  access_str = "Designer ";
    else if (access_level == ACL_LEVEL_MANAGER)   access_str = "Manager  ";
    else                                          access_str = "???      ";

    printf("  %s  \"%s\"\n", access_str, entry_name);
}


#if defined(__BORLANDC__)
#  pragma argsused
#endif

STATUS dump_acl_by_dbhandle(HANDLE db, ITEM_TABLE *summary, void *args)
{
    STATUS          status;
    HANDLE          acl;
    char            roles[ACL_PRIVCOUNT][ACL_PRIVSTRINGMAX];
    char            dbpath[MAXPATH];
    int             i;

    RETURN_IF_ERROR("EditACL.NSFDbPathGet", NSFDbPathGet(db, NULL, dbpath));
    printf("%s\n", dbpath);

    IF_ERROR("DumpACL.NSFDbReadACL", NSFDbReadACL(db, &acl))
        goto bad_acl;

    IF_ERROR("DumpACL.GetRoles", acl_get_roles(acl, roles))
        goto bad_roles;

    IF_ERROR("DumpACL.ACLEnumEntries", ACLEnumEntries(acl, dump_acl_entry, roles))
        goto bad_enum;

    printf("\n");

bad_enum:
bad_roles:
    OSMemFree(acl);
bad_acl:
    return 0;  //don't break enumeration by 'status';
}


STATUS dump_acl_by_db_filename(char *db_filename)
{
    STATUS status = 0;
    do_action_for_each_database(db_filename, server_name, dump_acl_by_dbhandle, NULL);
    return status;
}


bool isSwitch(char *str) {
    switch(*str) { case '!': case '-': case '+': case '=': return true; }
    return false;
}


int syntax(char *error_message)
{
    if (error_message)
        fprintf(STDERR, "%s\n", error_message);
    fprintf(STDERR,
        "LNACLs v2.0. Written by Ilya Evseev (evseev@csa.ru) at Feb 2001\n"
        "\n"
        "Dump and edit ACL (Access Control List) of Lotus Notes databases.\n"
        "\n"
        "Syntax: lnacl [:server] files.. acl.. files.. [acl..]    where acl is in form:\n"
        "\n"
        "  [!]{+|-|=}[UserName]:[UserType]:[Level]:[Modifiers]:[Roles]\n"
        "\n"
        "        !    delete all entries from database ACL before editing\n"
        "        +    add attribs to existing entry (or create new entry)\n"
        "        -    remove attribs from existing entry, or completely remove entry\n"
        "        =    replace existing entry by new attribs (or create new entry)\n"
        "\n"
        "  UserName:  CN=First Middle Last/O=Organization/C=Country (abbreviated ok too)\n"
        "  UserType:  {Person|Server|Group|PGroup|SGroup|Unknown}\n"
        "  Level:     {NoAccess|Depositor|Reader|Author|Designer|Manager}\n"
        "\n"
        "  Modifiers: cd = Create documents\n"
        "             dd = Delete documents\n"
        "             cpa = Create personal agents\n"
        "             cpf = Create personal folders/views\n"
        "             csf = Create shared folders/views\n"
        "             cls = Create LotusScript/Java agents\n"
        "             rpd = Read public documents\n"
        "             wpd = Write public documents\n"
        "\n"
        "Usage notes:\n"
        "\n"
        " * if cmdline doesn't contains filenames, filenames are be read from STDIN\n"
        "\n"
        " * if cmdline doesn't contains ACL's, database ACL will be dumped\n"
        "\n"
        " * filenames without path are lookups in Notes data directory\n"
        "\n"
        " * multiple modifiers and roles in one ACL are separated by \"+\"\n"
        "\n"
        " * each part of ACL may be empty:\n"
        "    - empty UserName is treated as \"-Default-\".\n"
        "    - all other parts be empty are treated as \"don't change this part\"\n"
        "\n"
        " * arguments with spaces are quoted by \"\"\n"
        "\n"
        " * UserType, Level and Modifiers are case-insensetive\n"
        "\n"
        " * UserType and Level are valid in shorten form\n"
        "   (no = NoAccess, u = Unknown, pg = PGroup, des = Designer, etc.)\n"
        "\n"
        "Environment variables:\n"
        "\n"
        "  * NO_ROLES - set/unset to simplify/restore dump format\n"
        "\n"
        "  * DOMINO_SERVER=name - set/unset for using local or remote server\n"
        "\n"
        "Examples:\n"
        "\n"
        " * Dump ACL of main Notes address book:\n"
        "   lnacls names\n"
        "\n"
        " * Create new entry or add attribs to existing entry:\n"
        "   lnacls names +NotesSrv/UCVT/RU:server:author:cpa+cpf:GroupCreator\n"
        "\n"
        " * Replace existing entry or create new:\n"
        "   lnacls names =NotesSrv/UCVT/RU:sgroup:ed::UserCreator+UserModifier\n"
        "\n"
        " * Remove attribs from existing entry:\n"
        "   lnacls names -NotesSrv/UCVT/RU:::cpf:UserCreator\n"
        "\n"
        " * Completely remove existing entry from NAB and LOG databases,\n"
        "   add new entry to both, dump results:\n"
        "   lnacls names log -NotesSrv/UCVT/RU +NotesSrv2/UCVT/RU:serv:reader names log\n"
        "\n"
        "try \"lnacls | more\", if output is too long for your display ;)))\n"
        );
    return 1;
}


int main(int argc, char *argv[])
{
    STATUS status;

    int firstFilenameIndex = 1;
    int firstSwitchIndex = -1;
    int currIndex = 1;
    int i;

#if 0
    {	int i;
	for (i = 0; i < argc; i++) {
	    printf("DEBUG: argv[%d] = \"%s\"\n", i, argv[i]);
	}
	return 0;
    }
#endif

    if (argc < 2)
        return syntax(NULL);

    if (argv[1][0] == ':') {
        server_name = &argv[1][1];
        currIndex++;
        firstFilenameIndex++;
    } else if ((server_name = getenv("DOMINO_SERVER")) == NULL) {
        server_name = "";
    }

    INIT_NOTES_SESSION();

    while (currIndex < argc) {
        if (!isSwitch(argv[currIndex])) {
            ++currIndex;
            continue;
        }
        firstSwitchIndex = currIndex;
        for (; (currIndex < argc) && isSwitch(argv[currIndex]);
            currIndex++);

        if (firstFilenameIndex == firstSwitchIndex) {
            char filename[1024];
            while (fgets(filename, sizeof(filename)-1, stdin)) {
                STATUS last = edit_acl_by_filename(filename, &argv[firstSwitchIndex],
                    currIndex - firstSwitchIndex);
                if (!status)
                    status = last;
            }
        } else {
            for (i = firstFilenameIndex; i < firstSwitchIndex; i++) {
                STATUS last = edit_acl_by_filename(argv[i], &argv[firstSwitchIndex],
                    currIndex - firstSwitchIndex);
                if (!status)
                    status = last;
            }
        }
        firstFilenameIndex = currIndex;
    }

    if (firstFilenameIndex > firstSwitchIndex)
        for (i = firstFilenameIndex; i < currIndex; i++) {
            STATUS last = dump_acl_by_db_filename(argv[i]);
            if (!status)
                status = last;
        }

    return (int)status;
}

/* EOF */
